"""Bookmarks storage manager - json

This file is a part of Bookmarks database and Internet robot.
"""

__author__ = "Oleg Broytman <phd@phdru.name>"
__copyright__ = "Copyright (C) 2010-2012 PhiloSoft Design"
__license__ = "GNU GPL"

__all__ = ['storage_json']


try:
   import json
except ImportError:
   import simplejson as json

from bkmk_objects import Folder, Bookmark, Ruler, Walker


class storage_json(Walker):
    filename = "bookmarks_db.json"

    def root_folder(self, f):
        self.dict = dict = {}
        dict["children"] = children = []
        self.folder_stack = [children]
        dict["dateAdded"] = convert_date_to_json(f.add_date)
        dict["id"] = f.id
        dict["lastModified"] = convert_date_to_json(f.last_modified)
        dict["root"] = "placesRoot"
        dict["title"] = ""
        dict["type"] = "text/x-moz-place-container"

    def start_folder(self, f, level):
        dict = {}
        comment = getattr(f, 'comment')
        if comment: dict["annos"] = make_annos(comment)
        dict["children"] = children = []
        dict["dateAdded"] = convert_date_to_json(f.add_date)
        dict["id"] = f.id
        index = getattr(f, 'index')
        if index: dict["index"] = index
        dict["lastModified"] = convert_date_to_json(f.last_modified)
        parent_idx = getattr(f, 'parent_idx')
        if parent_idx: dict["parent"] = parent_idx
        root = getattr(f, 'root')
        if root: dict["root"] = root
        dict["title"] = f.name.decode('utf-8')
        dict["type"] = "text/x-moz-place-container"
        self.folder_stack[-1].append(dict)
        self.folder_stack.append(children)

    def end_folder(self, f, level):
        del self.folder_stack[-1]

    def bookmark(self, b, level):
        dict = {}
        comment = getattr(b, 'comment')
        if comment: dict["annos"] = make_annos(comment)
        charset = getattr(b, 'charset')
        if charset: dict["charset"] = charset
        dict["dateAdded"] = convert_date_to_json(b.add_date)
        dict["id"] = b.id
        index = getattr(b, 'index')
        if index: dict["index"] = index
        keyword = getattr(b, 'keyword')
        if keyword: dict["keyword"] = keyword
        dict["lastModified"] = convert_date_to_json(b.last_modified)
        dict["parent"] = b.parent_idx
        dict["title"] = b.name.decode('utf-8')
        dict["type"] = "text/x-moz-place"
        dict["uri"] = uri = b.href
        if uri.startswith('place:'):
            if uri.startswith('place:folder'):
                value = 'Recently Bookmarked'
            elif uri.startswith('place:sort=8') or \
                    uri.startswith('place:redirectsMode'):
                value = 'Most Visited'
            elif uri.startswith('place:sort=14') or \
                    uri.startswith('place:type=6'):
                value = 'Recent Tags'
            else:
                raise ValueError('UNKNOWN place URI: %s' % uri)
            dict["annos"] = make_annos(value, name='Places/SmartBookmark')
            del dict["dateAdded"]
            del dict["lastModified"]
        self.folder_stack[-1].append(dict)

    def ruler(self, r, level):
        dict = {}
        comment = getattr(r, 'comment')
        if comment: dict["annos"] = make_annos(comment)
        dict["dateAdded"] = convert_date_to_json(r.add_date)
        dict["id"] = r.id
        dict["index"] = r.index
        dict["lastModified"] = convert_date_to_json(r.last_modified)
        dict["parent"] = r.parent_idx
        dict["title"] = r.name.decode('utf-8')
        dict["type"] = "text/x-moz-place-separator"
        self.folder_stack[-1].append(dict)

    def store(self, root_folder):
        root_folder.walk_depth(self)

        outfile = open(self.filename, 'wb')
        json.dump(self.dict, outfile)
        outfile.close()
        del self.dict

    def load(self):
        infile = open(self.filename, 'rb')
        bkmk_s = infile.read()
        infile.close()

        # Work around a bug in Mozilla - remove the trailing comma
        bkmk_s = bkmk_s.strip().replace(',]', ']')
        bookmarks_dict = json.loads(bkmk_s)
        del bkmk_s

        root_folder = Folder()
        root_folder.header = ''
        root_folder.add_date = convert_date_from_json(bookmarks_dict["dateAdded"])
        root_folder.comment = ''
        root_folder.last_modified = convert_date_from_json(bookmarks_dict["lastModified"])
        self.folder_stack = [root_folder]
        self.current_folder = root_folder

        self.load_folder(root_folder, bookmarks_dict)
        if self.folder_stack:
            raise RuntimeError('Excessive folder stack: %s' % self.folder_stack)

        return root_folder

    def load_folder(self, folder, fdict):
        if fdict["type"] != "text/x-moz-place-container":
            raise ValueError("The object is not a Mozilla container")

        folder.id = fdict["id"]
        folder.index = fdict.get("index")
        folder.parent_idx = fdict.get("parent")
        folder.root = fdict.get("root")
        folder.name = encode_title(fdict["title"])

        for record in fdict["children"]:
            if record["type"] == "text/x-moz-place-container":
                folder = Folder(
                    add_date=convert_date_from_json(record["dateAdded"]),
                    comment=get_comment(record.get("annos")),
                    last_modified=convert_date_from_json(record["lastModified"]))
                self.current_folder.append(folder)
                self.folder_stack.append(folder)
                self.current_folder = folder
                self.load_folder(folder, record)

            elif record["type"] == "text/x-moz-place":
                bookmark = Bookmark(
                    href=record["uri"].encode('utf-8'),
                    add_date=convert_date_from_json(record.get("dateAdded")),
                    last_modified=convert_date_from_json(record.get("lastModified")),
                    keyword=get_str(record, "keyword"),
                    comment=get_comment(record.get("annos")),
                    charset=get_str(record, "charset"))
                bookmark.id = record["id"]
                bookmark.index = record.get("index")
                bookmark.parent_idx = record["parent"]
                bookmark.name = encode_title(record["title"])
                self.current_folder.append(bookmark)

            elif record["type"] == "text/x-moz-place-separator":
                ruler = Ruler()
                ruler.add_date = convert_date_from_json(record["dateAdded"])
                ruler.id = record["id"]
                ruler.index = record["index"]
                ruler.last_modified = convert_date_from_json(record["lastModified"])
                ruler.parent_idx = record["parent"]
                ruler.name = encode_title(record["title"])
                ruler.comment = get_comment(record.get("annos"))
                self.current_folder.append(ruler)

            else:
                raise ValueError('Unknown record type "%s"' % record["type"])

        del self.folder_stack[-1]
        if self.folder_stack:
            self.current_folder = self.folder_stack[-1]
        else:
            self.current_folder = None


def convert_date_to_json(date):
    if date:
        date = int(date) * 10**6
    return date

def convert_date_from_json(date):
    if date:
        date = int(date)
        if date > 10**10:
            date /= 10**6
    return date

def encode_title(title):
    return title.encode("UTF-8", "xmlcharrefreplace")

def get_str(record, name):
    if name in record:
        return record[name].encode('utf-8')
    return ''

def get_comment(annos):
    if not annos:
        return ''

    for a in annos:
        if a["name"] == "bookmarkProperties/description":
            return a["value"].encode('utf-8')

    return ''

def make_annos(value, name="bookmarkProperties/description"):
    return [{
        "expires": 4,
        "flags": 0,
        "mimeType": None,
        "name": name,
        "type": 3,
        "value": value.decode('utf-8')}]
